package java.util.stream;

import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountedCompleter;
import java.util.Spliterator;
import java.util.function.DoubleConsumer;
import java.util.function.LongConsumer;
import java.util.function.IntConsumer;
import java.util.Objects;
import java.util.function.Consumer;

final class ForEachOps
{
    private ForEachOps() {
    }
    
    public static <T> TerminalOp<T, Void> makeRef(Consumer<? super T> obj, boolean b) {
        Objects.requireNonNull(obj);
        return new ForEachOp.OfRef(obj, b);
    }
    
    public static TerminalOp<Integer, Void> makeInt(IntConsumer obj, boolean b) {
        Objects.requireNonNull(obj);
        return new ForEachOp.OfInt(obj, b);
    }
    
    public static TerminalOp<Long, Void> makeLong(LongConsumer obj, boolean b) {
        Objects.requireNonNull(obj);
        return new ForEachOp.OfLong(obj, b);
    }
    
    public static TerminalOp<Double, Void> makeDouble(DoubleConsumer obj, boolean b) {
        Objects.requireNonNull(obj);
        return new ForEachOp.OfDouble(obj, b);
    }
    
    abstract static class ForEachOp<T> implements TerminalOp<T, Void>, TerminalSink<T, Void>
    {
        private final boolean ordered;
        
        protected ForEachOp(boolean ordered) {
            this.ordered = ordered;
        }
        
        @Override
        public int getOpFlags() {
            return this.ordered ? 0 : StreamOpFlag.NOT_ORDERED;
        }
        
        @Override
        public <S> Void evaluateSequential(PipelineHelper<T> pipelineHelper, Spliterator<S> spliterator) {
            return ((ForEachOps.ForEachOp)pipelineHelper.wrapAndCopyInto(this, spliterator)).get();
        }
        
        @Override
        public <S> Void evaluateParallel(PipelineHelper<T> pipelineHelper, Spliterator<S> spliterator) {
            if (this.ordered) {
                new ForEachOps.ForEachOrderedTask(pipelineHelper, spliterator, this).invoke();
            }
            else {
                new ForEachOps.ForEachTask(pipelineHelper, spliterator, pipelineHelper.wrapSink(this)).invoke();
            }
            return null;
        }
        
        @Override
        public Void get() {
            return null;
        }
        
        static final class OfRef<T> extends ForEachOps.ForEachOp<T>
        {
            final Consumer<? super T> consumer;
            
            OfRef(Consumer<? super T> consumer, boolean b) {
                super(b);
                this.consumer = consumer;
            }
            
            @Override
            public void accept(T t) {
                this.consumer.accept(t);
            }
        }
        
        static final class OfInt extends ForEachOps.ForEachOp<Integer> implements Sink.OfInt
        {
            final IntConsumer consumer;
            
            OfInt(IntConsumer consumer, boolean b) {
                super(b);
                this.consumer = consumer;
            }
            
            @Override
            public StreamShape inputShape() {
                return StreamShape.INT_VALUE;
            }
            
            @Override
            public void accept(int n) {
                this.consumer.accept(n);
            }
        }
        
        static final class OfLong extends ForEachOps.ForEachOp<Long> implements Sink.OfLong
        {
            final LongConsumer consumer;
            
            OfLong(LongConsumer consumer, boolean b) {
                super(b);
                this.consumer = consumer;
            }
            
            @Override
            public StreamShape inputShape() {
                return StreamShape.LONG_VALUE;
            }
            
            @Override
            public void accept(long n) {
                this.consumer.accept(n);
            }
        }
        
        static final class OfDouble extends ForEachOps.ForEachOp<Double> implements Sink.OfDouble
        {
            final DoubleConsumer consumer;
            
            OfDouble(DoubleConsumer consumer, boolean b) {
                super(b);
                this.consumer = consumer;
            }
            
            @Override
            public StreamShape inputShape() {
                return StreamShape.DOUBLE_VALUE;
            }
            
            @Override
            public void accept(double n) {
                this.consumer.accept(n);
            }
        }
    }
    
    static final class ForEachTask<S, T> extends CountedCompleter<Void>
    {
        private Spliterator<S> spliterator;
        private final Sink<S> sink;
        private final PipelineHelper<T> helper;
        private long targetSize;
        
        ForEachTask(PipelineHelper<T> helper, Spliterator<S> spliterator, Sink<S> sink) {
            super(null);
            this.sink = sink;
            this.helper = helper;
            this.spliterator = spliterator;
            this.targetSize = 0L;
        }
        
        ForEachTask(ForEachOps.ForEachTask<S, T> completer, Spliterator<S> spliterator) {
            super(completer);
            this.spliterator = spliterator;
            this.sink = completer.sink;
            this.targetSize = completer.targetSize;
            this.helper = completer.helper;
        }
        
        @Override
        public void compute() {
            Spliterator<Object> spliterator = this.spliterator;
            long n = spliterator.estimateSize();
            long targetSize;
            if ((targetSize = this.targetSize) == 0L) {
                targetSize = (this.targetSize = AbstractTask.suggestTargetSize(n));
            }
            int known = StreamOpFlag.SHORT_CIRCUIT.isKnown(this.helper.getStreamAndOpFlags());
            int n2 = 0;
            Sink<Object> sink = this.sink;
            while (known == 0 || !sink.cancellationRequested()) {
                Spliterator<Object> trySplit;
                if (n <= targetSize || (trySplit = spliterator.trySplit()) == null) {
                    this.helper.copyInto(sink, spliterator);
                    break;
                }
                ForEachOps.ForEachTask o = new ForEachOps.ForEachTask(this, trySplit);
                this.addToPendingCount(1);
                if (n2 != 0) {
                    n2 = 0;
                    spliterator = trySplit;
                    this = o;
                }
                else {
                    n2 = 1;
                    this = o;
                }
                this.fork();
                n = spliterator.estimateSize();
            }
            this.spliterator = null;
            this.propagateCompletion();
        }
    }
    
    static final class ForEachOrderedTask<S, T> extends CountedCompleter<Void>
    {
        private final PipelineHelper<T> helper;
        private Spliterator<S> spliterator;
        private final long targetSize;
        private final ConcurrentHashMap<ForEachOps.ForEachOrderedTask<S, T>, ForEachOps.ForEachOrderedTask<S, T>> completionMap;
        private final Sink<T> action;
        private final ForEachOps.ForEachOrderedTask<S, T> leftPredecessor;
        private Node<T> node;
        
        protected ForEachOrderedTask(PipelineHelper<T> helper, Spliterator<S> spliterator, Sink<T> action) {
            super(null);
            this.helper = helper;
            this.spliterator = spliterator;
            this.targetSize = AbstractTask.suggestTargetSize(spliterator.estimateSize());
            this.completionMap = new ConcurrentHashMap(Math.max(16, AbstractTask.LEAF_TARGET << 1));
            this.action = action;
            this.leftPredecessor = null;
        }
        
        ForEachOrderedTask(ForEachOps.ForEachOrderedTask<S, T> completer, Spliterator<S> spliterator, ForEachOps.ForEachOrderedTask<S, T> leftPredecessor) {
            super(completer);
            this.helper = completer.helper;
            this.spliterator = spliterator;
            this.targetSize = completer.targetSize;
            this.completionMap = completer.completionMap;
            this.action = completer.action;
            this.leftPredecessor = leftPredecessor;
        }
        
        @Override
        public final void compute() {
            doCompute(this);
        }
        
        private static <S, T> void doCompute(ForEachOps.ForEachOrderedTask<S, T> oldValue) {
            Spliterator<Object> spliterator = oldValue.spliterator;
            long targetSize = oldValue.targetSize;
            int n = 0;
            Spliterator<Object> trySplit;
            while (spliterator.estimateSize() > targetSize && (trySplit = spliterator.trySplit()) != null) {
                ForEachOps.ForEachOrderedTask o = new ForEachOps.ForEachOrderedTask<Object, Object>(oldValue, trySplit, oldValue.leftPredecessor);
                ForEachOps.ForEachOrderedTask value = new ForEachOps.ForEachOrderedTask(oldValue, spliterator, o);
                oldValue.addToPendingCount(1);
                value.addToPendingCount(1);
                oldValue.completionMap.put(o, value);
                if (oldValue.leftPredecessor != null) {
                    o.addToPendingCount(1);
                    if (oldValue.completionMap.replace(oldValue.leftPredecessor, oldValue, o)) {
                        oldValue.addToPendingCount(-1);
                    }
                    else {
                        o.addToPendingCount(-1);
                    }
                }
                if (n != 0) {
                    n = 0;
                    spliterator = trySplit;
                    oldValue = o;
                }
                else {
                    n = 1;
                    oldValue = value;
                    value = o;
                }
                value.fork();
            }
            if (oldValue.getPendingCount() > 0) {
                oldValue.node = ((Node.Builder)oldValue.helper.wrapAndCopyInto(oldValue.helper.makeNodeBuilder(oldValue.helper.exactOutputSizeIfKnown(spliterator), n2/* int, */ -> (Object[])new Object[n2]), spliterator)).build();
                oldValue.spliterator = null;
            }
            oldValue.tryComplete();
        }
        
        @Override
        public void onCompletion(CountedCompleter<?> countedCompleter) {
            if (this.node != null) {
                this.node.forEach(this.action);
                this.node = null;
            }
            else if (this.spliterator != null) {
                this.helper.wrapAndCopyInto(this.action, this.spliterator);
                this.spliterator = null;
            }
            ForEachOps.ForEachOrderedTask<Object, Object> forEachOrderedTask = (ForEachOps.ForEachOrderedTask)this.completionMap.remove(this);
            if (forEachOrderedTask != null) {
                forEachOrderedTask.tryComplete();
            }
        }
    }
}
